bl_info = {
    "name": "Corner Sharpener",
    "author": "JohnBogle",
    "version": (1, 0),
    "blender": (4, 1, 1),
    "location": "View3D > UI > Tool",
    "description": "The add-on sharpens the corners of a model under the subdivision modifier. The add-on is available when at least one vertex of an object in edit mode is selected. Ensure to inset the face before using the add-on. For the tool to sharpen a corner vertex, it needs to be connected to two boundary edges and one face edge in the mesh.",
    "warning": "",
    "doc_url": "",
    "category": "Tool",
}

import bpy
import bmesh
from bpy.types import Operator,Panel

def find_vertex_index_by_coordinates(target_co):

    obj = bpy.context.object
    
    # Get the BMesh data
    bm = bmesh.from_edit_mesh(obj.data)
    
    # Ensure the lookup table is up-to-date
    bm.verts.ensure_lookup_table()
    
    # Iterate through vertices to find the one with matching coordinates
    for v in bm.verts:
        if v.co == target_co:
            return v.index
        
    return None

class OBJECT_OT_move_vertices(Operator):
    bl_idname = "mesh.move_vertices"
    bl_label = "Sharp It"
    bl_options = {'REGISTER', 'UNDO'}

    relative_distance: bpy.props.FloatProperty(
                        name="Factor",
                        description="Define the slide size",
                        default=0.5,
                        min = -1.0,
                        max =1.0,
                    )
    
    even: bpy.props.BoolProperty(
        name="Even",
        description="Check this box to apply the effect evenly",
        default=True,
    )
    
    def subdivide_and_connect_faces(self):
        obj = bpy.context.object
        
        bm = bmesh.from_edit_mesh(obj.data)
        
        selected_verts_co = [v.co for v in bm.verts if v.select]
        
        to_slide_verts = []
        
        for vert_co in selected_verts_co:
        
            bm = bmesh.from_edit_mesh(obj.data)
            
            v_index = find_vertex_index_by_coordinates(vert_co)
            vert = bm.verts[v_index]
            
            edges = [edge for edge in vert.link_edges]
            
            #separating boundary_edges and face edges
            boundary_edges = [edge for edge in edges if edge.is_boundary]
            face_edges = [edge for edge in edges if edge not in boundary_edges]

            #It skips the vertex if the length of boundary edges are not 2 and face edge is not 1\
            if len(boundary_edges) != 2 or len(face_edges) != 1:
                continue
            
            #subdividing_boundary_edges
            bmesh.ops.subdivide_edges(bm, edges=boundary_edges, cuts=1)
            
            bmesh.update_edit_mesh(obj.data)
            
            bm = bmesh.from_edit_mesh(obj.data)
            
            bm.verts.ensure_lookup_table()
            
            vert = bm.verts[v_index]
            
            
            connected_verts = [edge.other_vert(vert) for edge in vert.link_edges]
            boundary_verts = [edge.other_vert(vert) for edge in vert.link_edges if edge.is_boundary]
            face_edge = [edge for edge in vert.link_edges if not edge.is_boundary]
            face_vert = [vert for vert in connected_verts if vert not in boundary_verts][0]


            for bvert in boundary_verts:
                bmesh.ops.connect_vert_pair(bm, verts=[face_vert, bvert])
                
            bmesh.ops.dissolve_edges(bm, edges=face_edge)
               
            bmesh.update_edit_mesh(obj.data)
            
            for b in boundary_verts:
                to_slide_verts.append((b.co, vert.co))

        to_slide_verts = [(find_vertex_index_by_coordinates(x),find_vertex_index_by_coordinates(y)) for x,y in to_slide_verts]

        return to_slide_verts

    def slide_verts(self,to_slide_verts):
        obj = bpy.context.object

        for a_index,b_index  in to_slide_verts:

            # Ensure we are using BMesh to handle mesh data
            bm = bmesh.from_edit_mesh(obj.data)

            bm.verts.ensure_lookup_table()
            
            # Get the vertices
            v_a = bm.verts[a_index]
            v_b = bm.verts[b_index]
            
            if not self.even:
                # Calculate the direction vector from A to B
                full_distance_vector = v_b.co - v_a.co
                
                relative_distance_vector = full_distance_vector * self.relative_distance
                
                # Move vertex A towards vertex B
                v_a.co += relative_distance_vector
            else:
                direction_vector = (v_b.co - v_a.co).normalized()

                # Calculate the new position for v_a such that the distance between v_a and v_b is (1 - self.relative_distance)
                v_a.co = v_b.co - direction_vector * (1 - self.relative_distance)
        
            
            # Update the mesh
            bmesh.update_edit_mesh(obj.data)
        
        return {'FINISHED'}
                   
    def execute(self,context):
        to_slide_verts = self.subdivide_and_connect_faces()
        
        if to_slide_verts:
            self.slide_verts(to_slide_verts=to_slide_verts)
        else:
            self.report({'ERROR'}, f"For the tool to sharpen a corner vertex, it needs to be connected to two boundary edges and one face edge in the mesh")
        return {'FINISHED'}
    
class VIEW3D_PT_move_vertices_panel(Panel):
    bl_label = "SharpCorner Tool"
    bl_idname = "VIEW3D_PT_move_vertices_panel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Tool"

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.operator("mesh.move_vertices", text="Sharpen It")

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return context.mode in {"EDIT_MESH"} and obj is not None and obj.type  == "MESH" and obj.data.total_vert_sel
       
def register():
    bpy.utils.register_class(VIEW3D_PT_move_vertices_panel)
    bpy.utils.register_class(OBJECT_OT_move_vertices)

def unregister():
    bpy.utils.unregister_class(VIEW3D_PT_move_vertices_panel)
    bpy.utils.unregister_class(OBJECT_OT_move_vertices)

if __name__ == "__main__":
    register()

    